{ =========================================================================== }
{                                 JORDAN.PAS                                  }
{ =========================================================================== }

unit Jordan;

{ ---> funcoes e procedimentos usados no calculo da forma canonica de Jordan

       (C) 1990, 1991 Lenimar Nunes de Andrade
           CCENDM03@BRUFPB.BITNET                                             }

{$E+,N+,F+,B-,V-,S+,R+}

interface

uses
  Equacao, Matrizes, FPTCBas, AlgLin, Printer;

type
  bloco = record
    lambda: real;                  { autovalor real                           }
    QuantBlocosLambda: byte;       { quantidade de blocos associados a lambda }
    ordem: array[1..MaxM] of byte; { ordem[i] = quantidade de blocos de ordem }
  end;                             {            i associados a lambda         }

  DBlocos = record
    a: array[1..MaxM] of bloco;    { blocos de autovalores reais. Cada a[i]   }
                                   { corresponde a um autovalor diferente     }
    QuantTotalBlocos: byte;        { total de blocos                          }
    QuantAutovalores: byte;        { quantidade de autovalores reais          }
  end;

procedure CalculaFormaDeJordan(mat: matriz; w: autovalores; min: minimo;
                                var HouveErro: boolean; var diagonal: DBlocos);
procedure FormataMatrizJordan(diagonal: DBlocos; var FJordan: matriz;
                                 OrdemMat: byte; w:  autovalores; min: minimo);
procedure CalculaBaseFJordan(mat: matriz; diagonal: DBlocos; min: minimo;
                                     var base: matriz; var AchouBase: boolean);
procedure ImprimeBase(base: matriz; op: InfoPress);

{ =========================================================================== }

implementation

const
  erro3: real = 1E-06; { usado no calculo da altura de um vetor }

{ --------------------------------------------------------------------------- }

procedure FormataBlocosComplexos(var mat: matriz; inicio, ordem: byte;
                                                                 re, im: real);

{ ---> formata bloco de Jordan associado a autovalor complexo de uma matriz
       nula mat a partir de uma posicao inicial na matriz                     }

var
  i, j: byte;

begin
  for i := 1 to 2*ordem do
  begin
    mat.a[i + inicio - 1, i + inicio - 1] := re;
    if (i < 2*ordem - 1) then mat.a[i + inicio  - 1, i + inicio + 1] := 1;
    if Odd(i) then
      mat.a[i + inicio, i + inicio - 1] := im
    else
      mat.a[i + inicio - 2, i + inicio - 1] := -im;
  end;
end;

{ --------------------------------------------------------------------------- }

function H(vet: vetor; aux: matriz): shortint;

var
  ColNula: boolean;
  i, j, k: byte;
  soma: real;
  aux2: matriz;

begin
  aux2 := aux;
  k := 1;
  repeat
    ColNula := true;
    i := 1;
    repeat
      soma := 0;
      for j :=  1 to aux.n do
        soma := soma + aux.a[i, j]*vet[j];
      if (Abs(soma) > erro3) then ColNula := false;
      Inc(i);
    until ((i > aux.m) or (not ColNula));
    if (not ColNula) then
    begin
      Inc(k);
      if (k > aux.m) then
      begin
        H := -1;
        Exit;
      end;
      MultiplicaMatrizes(aux, aux2, aux);
    end;
  until ColNula;
  H := k;
end;

{ --------------------------------------------------------------------------- }

procedure CalculaFormaDeJordan(mat: matriz; w: autovalores; min: minimo;
                                var HouveErro: boolean; var diagonal: DBlocos);

var
  i, j, k: byte;
  m: array[0..MaxM] of byte;

begin
  HouveErro := false;
  if (w.QuantComplex > 0) then
    for i := 1 to w.QuantComplex do
      if (w.AutovalorCx[i].mult - min.expo[i + w.QuantRe] >= 2) then
        HouveErro := true;
  if HouveErro then Exit;
  diagonal.QuantAutovalores :=  w.QuantRe;
  diagonal.QuantTotalBlocos := 0;
  for i := 1 to w.QuantRe do
  begin
    diagonal.a[i].lambda := w.AutovalorRe[i].x;
    diagonal.a[i].QuantBlocosLambda := Nulidade(mat, diagonal.a[i].lambda, 1);
    Inc(diagonal.QuantTotalBlocos, diagonal.a[i].QuantBlocosLambda);
    for k := 1 to min.expo[i] do
      m[k] := Nulidade(mat, diagonal.a[i].lambda, k);
    m[0] := 0;
    diagonal.a[i].ordem[min.expo[i]] := m[min.expo[i]] - m[min.expo[i] - 1];
    for k := 1 to (min.expo[i] - 1) do
      diagonal.a[i].ordem[k] := 2*m[k] - m[k + 1]  - m[k - 1];
  end;
end;

{ --------------------------------------------------------------------------- }

procedure FormataMatrizJordan(diagonal: DBlocos; var FJordan: matriz;
                                  OrdemMat: byte; w: autovalores; min: minimo);

var
  ExpoCar, ExpoMin, inicio, i, j, k, l, deslocamento: byte;

begin
  FJordan.m := OrdemMat; FJordan.n := OrdemMat;
  for i := 1 to FJordan.m do
    for j := 1 to FJordan.n do
      FJordan.a[i, j] := 0;
  deslocamento := 0;
  for i := 1 to w.QuantRe do
  begin
    for j := min.expo[i] downto 1 do
    begin
      for k := 1 to diagonal.a[i].ordem[j] do
      begin
        for l := 1 to j do
        begin
          FJordan.a[l + deslocamento, l + deslocamento] :=
                                                          diagonal.a[i].lambda;
          if (l < j) then
            FJordan.a[l + deslocamento, l + 1 + deslocamento] := 1;
        end;
        Inc(deslocamento, j);
      end;
    end;
  end;
  if (w.QuantComplex = 0) then Exit;
  inicio := 1;
  for i := 1 to w.QuantRe do
    for j := 1 to w.AutovalorRe[i].mult do
      Inc(inicio);
  for i := 1 to w.QuantComplex do
  begin
    ExpoCar := w.AutovalorCx[i].mult;
    ExpoMin := min.expo[i + w.QuantRe];
    if (ExpoCar - ExpoMin >= 2) then Exit;
    if (ExpoCar = ExpoMin) then
    begin
      FormataBlocosComplexos(FJordan, inicio, ExpoMin, w.AutovalorCx[i].a,
                                                           w.AutovalorCx[i].b);
      Inc(inicio, 2*ExpoMin);
    end
    else
    begin
      FormataBlocosComplexos(FJordan, inicio, ExpoMin, w.AutovalorCx[i].a,
                                                           w.AutovalorCx[i].b);
      Inc(inicio, 2*ExpoMin);
      FormataBlocosComplexos(FJordan, inicio, 1, w.AutovalorCx[i].a,
                                                           w.AutovalorCx[i].b);
      Inc(inicio, 2);
    end;
  end;
end;

{ --------------------------------------------------------------------------- }

procedure CalculaBaseFJordan(mat: matriz; diagonal: DBlocos; min: minimo;
                                     var base: matriz; var AchouBase: boolean);

var
  PriOpcaoMenor, UltOpcaoMenor, PrimeiraOpcao, UltimaOpcao, selecao, qb, q,
                                               x, y, linha, col, i, j, k: byte;
  VetBasEncontrados: array[0..MaxM] of byte;
  sol: resposta;
  aux, aux2: matriz;
  vet: vetor;
  altura: shortint;
  VetoresDeAlturaH: matriz;
  XAux: real;
  BaseLI: boolean;

begin
  aux := mat; AchouBase := true; base.m := 0; base.n := mat.n;
  VetBasEncontrados[0] := 0;
  for i := 1 to diagonal.QuantAutovalores do
  begin
    VetBasEncontrados[i] := VetBasEncontrados[i - 1];
    PrimeiraOpcao := 1; PriOpcaoMenor := 1;
    for j := min.expo[i] downto 1 do
      if (diagonal.a[i].ordem[j] > 0) then
      begin
        for linha := 1 to aux.m do
          aux.a[linha, linha] := mat.a[linha, linha] - diagonal.a[i].lambda;
        aux2 := aux;
        ElevaMatriz(aux, j);
        Inc(aux.n);
        for linha := 1 to aux.m do aux.a[linha, aux.n] := 0;
        ResolveSistema(aux, sol);
        Dec(aux.n);
        aux := aux2;
        q := 0;                       { determina os vetores de altura h = j }
        VetoresDeAlturaH.n := mat.n;
        for linha := 1 to sol.x.m do
        begin
          for col := 1 to sol.x.n do
            vet[col] := sol.x.a[linha, col];
            altura := H(vet, aux);
            if (altura = j) then
            begin
              Inc(q);
              for x := 1 to sol.x.n do
                VetoresDeAlturaH.a[q, x] := vet[x];
            end;
        end;
        VetoresDeAlturaH.m := q;
        if (diagonal.a[i].ordem[j] > q) then
        begin
          AchouBase := false;
          Exit;
        end;
        if (j = min.expo[i]) then
          UltimaOpcao := VetoresDeAlturaH.m
        else
          UltOpcaoMenor := VetoresDeAlturaH.m;
        qb := VetBasEncontrados[i];             { determina a base }
        for x := 1 to diagonal.a[i].ordem[j] do
        begin
          Inc(qb);
          base.m := qb;
          if (j = min.expo[i]) then
            selecao := PrimeiraOpcao + x - 1
          else
            selecao := PriOpcaoMenor + x - 1;
          repeat
            for y := 1 to mat.n do
              vet[y] := VetoresDeAlturaH.a[selecao, y];
            for y := 1 to mat.n do
              base.a[qb, y] := vet[y];
            BaseLI := LI(base);
            if (not BaseLI) then Inc(selecao);
          until (BaseLI or (selecao > VetoresDeAlturaH.m));
          if (selecao > VetoresDeAlturaH.m) then
          begin
            AchouBase := false;
            Exit;
          end;
          repeat
            if (j > 1) then
            begin
              Inc(qb); base.m := qb;
              for linha := 1 to aux.m do
              begin
                base.a[qb, linha]  := 0;
                for col := 1 to aux.n do
                  base.a[qb, linha] := base.a[qb, linha] +
                                                  vet[col]*aux.a[linha, col];
              end;
              for y := 1 to mat.n do vet[y] := base.a[qb, y];
            end;
          until ((qb - VetBasEncontrados[i]) mod j = 0);
          for y := (VetBasEncontrados[i] + 1) to
                                       (VetBasEncontrados[i] + Trunc(j/2)) do
            for col := 1 to mat.n do
            begin
              XAux := base.a[y, col];
              base.a[y, col] := base.a[qb + VetBasEncontrados[i] + 1-y, col];
              base.a[qb + VetBasEncontrados[i] + 1 - y, col] := XAux;
            end;
          VetBasEncontrados[i] := qb;
          if ((not LI(base)) and (j <> min.expo[i])) then
          begin
            if (PriOpcaoMenor < UltOpcaoMenor) then
              Inc(PriOpcaoMenor)
            else
              if (PrimeiraOpcao < UltimaOpcao) then
              begin
                PriOpcaoMenor := 1;
                Inc(PrimeiraOpcao);
              end
            else
            begin
              AchouBase := false;
              Exit;
            end;
            x := diagonal.a[i].ordem[j];
            j := min.expo[i] + 1;
            VetBasEncontrados[i] := VetBasEncontrados[i - 1];
          end;
        end;
      end;
  end;
  if LI(base) then AchouBase := true else AchouBase := false;
end;

{ --------------------------------------------------------------------------- }

procedure ImprimeBase(base: matriz; op: InfoPress);

var
  brancos, compr, i, j, k, tam1, tam2, imp, LinhasImpressas: byte;
  aux: matriz;
  ch: char;
  fim: boolean;

begin
  aux := base; fim := false; op.n := base.m; op.MaxL := 6;
  FormataMatriz(aux, tam1, tam2);
  if (op.racional <> 0) then ReformataMatriz(aux, tam1, tam2);
  compr := (tam1 + 1)*base.n;
  if (compr >= op.largura) then
    brancos := 1
  else
    brancos := (op.largura - compr) div 2;
  LinhasImpressas := 0;
  for j := 1 to aux.m do
  begin
    if (op.tela = 1) then
      Write(' ':brancos, '(')
    else
      Write(Lst, ' ':brancos, '(');
    imp := brancos + 1;
    for k := 1 to aux.n do
    begin
      if (op.tela = 1) then
        if (imp + tam1 + 1 <= op.largura) then
        begin
          if (op.racional = 0) then
            Write(aux.a[j, k]:tam1:tam2)
          else
            Write(Rac(aux.a[j, k], tam1, tam2));
          imp := imp + tam1 + 1;
        end
        else
        begin
          imp := brancos + 1;
          Writeln; Inc(LinhasImpressas);
          if (LinhasImpressas mod op.MaxL = 0) and (op.tela = 1) and
                                          (op.opcao = 1) and (base.m >= 6) then
            PausaJanMenu(' Base que fornece a forma de Jordan ');
          if (op.racional = 0) then
            Write(' ':imp, aux.a[j, k]:tam1:tam2)
          else
            Write(' ':imp, Rac(aux.a[j, k], tam1, tam2));
          imp := imp + tam1 + 1;
        end
      else
        if (imp + tam1 + 1 <= op.largura) then
        begin
          if (op.racional = 0) then
            Write(Lst, aux.a[j, k]:tam1:tam2)
          else
            Write(Lst, Rac(aux.a[j, k], tam1, tam2));
          imp := imp + tam1 + 1;
        end
        else
        begin
          imp := brancos + 1;
          Writeln(Lst);
          if (op.racional = 0) then
            Write(Lst, ' ':imp, aux.a[j, k]:tam1:tam2)
          else
            Write(Lst, ' ':imp, Rac(aux.a[j, k], tam1, tam2));
          imp := imp + tam1 + 1;
        end;
      if (k < aux.n) then
        if (op.tela = 1) then Write(',') else Write(Lst, ',');
    end;
    if (op.tela = 1) then Writeln(')') else Writeln(Lst, ')');
    fim := (j = aux.m);
    if not fim then
    begin
      Inc(LinhasImpressas);
      if (LinhasImpressas mod op.MaxL = 0) and (op.tela = 1) and
                                          (op.opcao = 1) and (base.m >= 6) then
        PausaJanMenu(' Base que fornece a forma de Jordan ');
    end;
  end;
end;

{ --------------------------------------------------------------------------- }

var
  arq: file of real;
  erro1, erro2: real;

begin
  Assign(arq, 'ALGLIN.ERR');
  {$I-}
  Reset(arq);
  {$I+}
  if (IOResult = 0) then
  begin
    Read(arq, erro1, erro2, erro3);
    Close(arq);
  end;
end.

{ =========================================================================== }

